@@ -3,7 +3,7 @@ package main
 import (
 	"errors"
 	"fmt"
-	"path/filepath"
+	"runtime/debug"
 	"strings"
 	"unicode"
 	"unicode/utf8"
@@ -19,16 +19,6 @@ func splitCFlagsFromArgs(in []string) (args, cflags []string) {
 	return in, nil
 }
 
-func splitPathAbs(path string) (dir, file string, err error) {
-	abs, err := filepath.Abs(path)
-	if err != nil {
-		return "", "", err
-	}
-
-	dir, file = filepath.Split(abs)
-	return
-}
-
 func splitArguments(in string) ([]string, error) {
 	var (
 		result  []string
@@ -92,3 +82,13 @@ func toUpperFirst(str string) string {
 	first, n := utf8.DecodeRuneInString(str)
 	return string(unicode.ToUpper(first)) + str[n:]
 }
+
+func currentModule() string {
+	bi, ok := debug.ReadBuildInfo()
+	if !ok {
+		// Fall back to constant since bazel doesn't support BuildInfo.
+		return "github.com/cilium/ebpf"
+	}
+
+	return bi.Main.Path
+}
