@@ -78,3 +78,34 @@ func TestIterMapElements(t *testing.T) {
 		t.Error("Non-empty output from no-op iterator:", string(contents))
 	}
 }
+
+func TestUDPIter(t *testing.T) {
+	// Introduced by 5788b3a07fc5 ("net: bpf: Implement bpf iterator for udp")
+	testutils.SkipOnOldKernel(t, "5.9", "udp iter")
+
+	prog := mustLoadProgram(t, ebpf.Tracing, ebpf.AttachTraceIter, "udp")
+
+	it, err := AttachIter(IterOptions{
+		Program: prog,
+	})
+	if err != nil {
+		t.Fatal("Can't create iter:", err)
+	}
+
+	file, err := it.Open()
+	if err != nil {
+		t.Fatal("Can't open iter instance:", err)
+	}
+	defer file.Close()
+
+	contents, err := io.ReadAll(file)
+	if err != nil {
+		t.Fatal(err)
+	}
+
+	if len(contents) != 0 {
+		t.Error("Non-empty output from no-op iterator:", string(contents))
+	}
+
+	testLink(t, it, prog)
+}
