@@ -12,7 +12,6 @@ import (
 	"unsafe"
 
 	"github.com/cilium/ebpf/asm"
-	"github.com/cilium/ebpf/btf"
 	"github.com/cilium/ebpf/internal"
 	"github.com/cilium/ebpf/internal/sys"
 	"github.com/cilium/ebpf/internal/testutils"
@@ -32,6 +31,21 @@ var (
 	}
 )
 
+// newHash returns a new Map of type Hash. Cleanup is handled automatically.
+func newHash(t *testing.T) *Map {
+	hash, err := NewMap(&MapSpec{
+		Type:       Hash,
+		KeySize:    5,
+		ValueSize:  4,
+		MaxEntries: 10,
+	})
+	if err != nil {
+		t.Fatal(err)
+	}
+	t.Cleanup(func() { hash.Close() })
+	return hash
+}
+
 func TestMap(t *testing.T) {
 	m := createArray(t)
 	defer m.Close()
@@ -382,20 +396,23 @@ func TestMapWithLock(t *testing.T) {
 			t.Fatalf("Want value 5, got %d", value.Cnt)
 		}
 
-		value.Cnt = 0
-		err = m.LookupAndDeleteWithFlags(&key, &value, LookupLock)
-		if err != nil {
-			t.Fatal(err)
-		}
-		if value.Cnt != 5 {
-			t.Fatalf("Want value 5, got %d", value.Cnt)
-		}
+		t.Run("LookupAndDelete", func(t *testing.T) {
+			testutils.SkipOnOldKernel(t, "5.14", "LOOKUP_AND_DELETE flags")
 
-		err = m.LookupWithFlags(&key, &value, LookupLock)
-		if err != nil && !errors.Is(err, ErrKeyNotExist) {
-			t.Fatal(err)
-		}
+			value.Cnt = 0
+			err = m.LookupAndDeleteWithFlags(&key, &value, LookupLock)
+			if err != nil {
+				t.Fatal(err)
+			}
+			if value.Cnt != 5 {
+				t.Fatalf("Want value 5, got %d", value.Cnt)
+			}
 
+			err = m.LookupWithFlags(&key, &value, LookupLock)
+			if err != nil && !errors.Is(err, ErrKeyNotExist) {
+				t.Fatal(err)
+			}
+		})
 	})
 }
 
@@ -619,6 +636,38 @@ func TestMapLoadPinned(t *testing.T) {
 	c.Assert(pinned, qt.IsTrue)
 }
 
+func TestMapLoadReusePinned(t *testing.T) {
+	c := qt.New(t)
+
+	for _, typ := range []MapType{Array, Hash, DevMap, DevMapHash} {
+		t.Run(typ.String(), func(t *testing.T) {
+			if typ == DevMap {
+				testutils.SkipOnOldKernel(t, "4.14", "devmap")
+			}
+			if typ == DevMapHash {
+				testutils.SkipOnOldKernel(t, "5.4", "devmap_hash")
+			}
+			tmp := testutils.TempBPFFS(t)
+			spec := &MapSpec{
+				Name:       "pinmap",
+				Type:       typ,
+				KeySize:    4,
+				ValueSize:  4,
+				MaxEntries: 1,
+				Pinning:    PinByName,
+			}
+
+			m1, err := NewMapWithOptions(spec, MapOptions{PinPath: tmp})
+			c.Assert(err, qt.IsNil)
+			defer m1.Close()
+
+			m2, err := NewMapWithOptions(spec, MapOptions{PinPath: tmp})
+			c.Assert(err, qt.IsNil)
+			defer m2.Close()
+		})
+	}
+}
+
 func TestMapLoadPinnedUnpin(t *testing.T) {
 	tmp := testutils.TempBPFFS(t)
 	c := qt.New(t)
@@ -859,7 +908,7 @@ func TestNewMapInMapFromFD(t *testing.T) {
 	defer nested.Close()
 
 	// Do not copy this, use Clone instead.
-	another, err := NewMapFromFD(nested.FD())
+	another, err := NewMapFromFD(dupFD(t, nested.FD()))
 	if err != nil {
 		t.Fatal("Can't create a new nested map from an FD")
 	}
@@ -1180,8 +1229,7 @@ func TestMapGuessNonExistentKey(t *testing.T) {
 }
 
 func TestNotExist(t *testing.T) {
-	hash := createHash()
-	defer hash.Close()
+	hash := newHash(t)
 
 	var tmp uint32
 	err := hash.Lookup("hello", &tmp)
@@ -1212,8 +1260,7 @@ func TestNotExist(t *testing.T) {
 }
 
 func TestExist(t *testing.T) {
-	hash := createHash()
-	defer hash.Close()
+	hash := newHash(t)
 
 	if err := hash.Put("hello", uint32(21)); err != nil {
 		t.Errorf("Failed to put key/value pair into hash: %v", err)
@@ -1492,18 +1539,12 @@ func TestMapFromFD(t *testing.T) {
 
 	// If you're thinking about copying this, don't. Use
 	// Clone() instead.
-	m2, err := NewMapFromFD(m.FD())
+	m2, err := NewMapFromFD(dupFD(t, m.FD()))
 	testutils.SkipIfNotSupported(t, err)
 	if err != nil {
 		t.Fatal(err)
 	}
-
-	// Both m and m2 refer to the same fd now. Closing either of them will
-	// release the fd to the OS, which then might re-use that fd for another
-	// test. Once we close the second map we might close the re-used fd
-	// inadvertently, leading to spurious test failures.
-	// To avoid this we have to "leak" one of the maps.
-	m2.fd.Forget()
+	defer m2.Close()
 
 	var val uint32
 	if err := m2.Lookup(uint32(0), &val); err != nil {
@@ -1577,8 +1618,8 @@ func TestMapGetNextID(t *testing.T) {
 	var next MapID
 	var err error
 
-	hash := createHash()
-	defer hash.Close()
+	// Ensure there is at least one map on the system.
+	_ = newHash(t)
 
 	if next, err = MapGetNextID(MapID(0)); err != nil {
 		t.Fatal("Can't get next ID:", err)
@@ -1604,8 +1645,7 @@ func TestMapGetNextID(t *testing.T) {
 }
 
 func TestNewMapFromID(t *testing.T) {
-	hash := createHash()
-	defer hash.Close()
+	hash := newHash(t)
 
 	info, err := hash.Info()
 	testutils.SkipIfNotSupported(t, err)
@@ -1652,15 +1692,13 @@ func TestMapPinning(t *testing.T) {
 	pinned := m1.IsPinned()
 	c.Assert(pinned, qt.IsTrue)
 
+	m1Info, err := m1.Info()
+	c.Assert(err, qt.IsNil)
+
 	if err := m1.Put(uint32(0), uint32(42)); err != nil {
 		t.Fatal("Can't write value:", err)
 	}
 
-	// This is a terrible hack: if loading a pinned map tries to load BTF,
-	// it will get a nil *btf.Spec from this *btf.Map. This is turn will make
-	// btf.NewHandle fail.
-	spec.BTF = new(btf.Spec)
-
 	m2, err := NewMapWithOptions(spec, MapOptions{PinPath: tmp})
 	testutils.SkipIfNotSupported(t, err)
 	if err != nil {
@@ -1668,6 +1706,14 @@ func TestMapPinning(t *testing.T) {
 	}
 	defer m2.Close()
 
+	m2Info, err := m2.Info()
+	c.Assert(err, qt.IsNil)
+
+	if m1ID, ok := m1Info.ID(); ok {
+		m2ID, _ := m2Info.ID()
+		c.Assert(m2ID, qt.Equals, m1ID)
+	}
+
 	var value uint32
 	if err := m2.Lookup(uint32(0), &value); err != nil {
 		t.Fatal("Can't read from map:", err)
@@ -1717,7 +1763,17 @@ func (cbv *customBenchValue) MarshalBinary() ([]byte, error) {
 	return buf, nil
 }
 
-func BenchmarkMarshalling(b *testing.B) {
+type benchKey struct {
+	id uint64
+}
+
+func (bk *benchKey) MarshalBinary() ([]byte, error) {
+	buf := make([]byte, 8)
+	internal.NativeEndian.PutUint64(buf, bk.id)
+	return buf, nil
+}
+
+func BenchmarkMarshaling(b *testing.B) {
 	newMap := func(valueSize uint32) *Map {
 		m, err := NewMap(&MapSpec{
 			Type:       Hash,
@@ -1737,8 +1793,9 @@ func BenchmarkMarshalling(b *testing.B) {
 	if err := m.Put(key, benchValue{}); err != nil {
 		b.Fatal(err)
 	}
+	b.Cleanup(func() { m.Close() })
 
-	b.Run("reflection", func(b *testing.B) {
+	b.Run("ValueUnmarshalReflect", func(b *testing.B) {
 		b.ReportAllocs()
 		b.ResetTimer()
 
@@ -1752,7 +1809,21 @@ func BenchmarkMarshalling(b *testing.B) {
 		}
 	})
 
-	b.Run("custom", func(b *testing.B) {
+	b.Run("KeyMarshalReflect", func(b *testing.B) {
+		b.ReportAllocs()
+		b.ResetTimer()
+
+		var value benchValue
+
+		for i := 0; i < b.N; i++ {
+			err := m.Lookup(&key, unsafe.Pointer(&value))
+			if err != nil {
+				b.Fatal("Can't get key:", err)
+			}
+		}
+	})
+
+	b.Run("ValueBinaryUnmarshaler", func(b *testing.B) {
 		b.ReportAllocs()
 		b.ResetTimer()
 
@@ -1766,7 +1837,22 @@ func BenchmarkMarshalling(b *testing.B) {
 		}
 	})
 
-	b.Run("unsafe", func(b *testing.B) {
+	b.Run("KeyBinaryMarshaler", func(b *testing.B) {
+		b.ReportAllocs()
+		b.ResetTimer()
+
+		var key benchKey
+		var value customBenchValue
+
+		for i := 0; i < b.N; i++ {
+			err := m.Lookup(&key, unsafe.Pointer(&value))
+			if err != nil {
+				b.Fatal("Can't get key:", err)
+			}
+		}
+	})
+
+	b.Run("KeyValueUnsafe", func(b *testing.B) {
 		b.ReportAllocs()
 		b.ResetTimer()
 
@@ -1802,6 +1888,7 @@ func BenchmarkPerCPUMarshalling(b *testing.B) {
 	if err := m.Put(key, val[0:]); err != nil {
 		b.Fatal(err)
 	}
+	b.Cleanup(func() { m.Close() })
 
 	b.Run("reflection", func(b *testing.B) {
 		b.ReportAllocs()
@@ -1828,6 +1915,7 @@ func BenchmarkMap(b *testing.B) {
 	if err != nil {
 		b.Fatal(err)
 	}
+	b.Cleanup(func() { m.Close() })
 
 	if err := m.Put(uint32(0), uint32(42)); err != nil {
 		b.Fatal(err)
@@ -1886,6 +1974,128 @@ func BenchmarkMap(b *testing.B) {
 	})
 }
 
+func BenchmarkIterate(b *testing.B) {
+	m, err := NewMap(&MapSpec{
+		Type:       Hash,
+		KeySize:    8,
+		ValueSize:  8,
+		MaxEntries: 1000,
+	})
+	if err != nil {
+		b.Fatal(err)
+	}
+	b.Cleanup(func() {
+		m.Close()
+	})
+
+	var (
+		n      = m.MaxEntries()
+		keys   = make([]uint64, n)
+		values = make([]uint64, n)
+	)
+
+	for i := 0; uint32(i) < n; i++ {
+		keys[i] = uint64(i)
+		values[i] = uint64(i)
+	}
+
+	if _, err := m.BatchUpdate(keys, values, nil); err != nil {
+		b.Fatal(err)
+	}
+
+	b.Run("MapIterator", func(b *testing.B) {
+		var k, v uint64
+
+		b.ReportAllocs()
+
+		for i := 0; i < b.N; i++ {
+			iter := m.Iterate()
+			for iter.Next(&k, &v) {
+				continue
+			}
+			if err := iter.Err(); err != nil {
+				b.Fatal(err)
+			}
+		}
+	})
+
+	b.Run("MapIteratorDelete", func(b *testing.B) {
+		var k, v uint64
+
+		b.ReportAllocs()
+
+		for i := 0; i < b.N; i++ {
+			b.StopTimer()
+			if _, err := m.BatchUpdate(keys, values, nil); err != nil {
+				b.Fatal(err)
+			}
+			b.StartTimer()
+
+			iter := m.Iterate()
+			for iter.Next(&k, &v) {
+				if err := m.Delete(&k); err != nil {
+					b.Fatal(err)
+				}
+			}
+			if err := iter.Err(); err != nil {
+				b.Fatal(err)
+			}
+		}
+	})
+
+	b.Run("BatchLookup", func(b *testing.B) {
+		k := make([]uint64, m.MaxEntries())
+		v := make([]uint64, m.MaxEntries())
+
+		b.ReportAllocs()
+
+		for i := 0; i < b.N; i++ {
+			var next uint32
+			_, err := m.BatchLookup(nil, &next, k, v, nil)
+			if err != nil && !errors.Is(err, ErrKeyNotExist) {
+				b.Fatal(err)
+			}
+		}
+	})
+
+	b.Run("BatchLookupAndDelete", func(b *testing.B) {
+		k := make([]uint64, m.MaxEntries())
+		v := make([]uint64, m.MaxEntries())
+
+		b.ReportAllocs()
+
+		for i := 0; i < b.N; i++ {
+			b.StopTimer()
+			if _, err := m.BatchUpdate(keys, values, nil); err != nil {
+				b.Fatal(err)
+			}
+			b.StartTimer()
+
+			var next uint32
+			_, err := m.BatchLookupAndDelete(nil, &next, k, v, nil)
+			if err != nil && !errors.Is(err, ErrKeyNotExist) {
+				b.Fatal(err)
+			}
+		}
+	})
+
+	b.Run("BatchDelete", func(b *testing.B) {
+		b.ReportAllocs()
+
+		for i := 0; i < b.N; i++ {
+			b.StopTimer()
+			if _, err := m.BatchUpdate(keys, values, nil); err != nil {
+				b.Fatal(err)
+			}
+			b.StartTimer()
+
+			if _, err := m.BatchDelete(keys, nil); err != nil {
+				b.Fatal(err)
+			}
+		}
+	})
+}
+
 // Per CPU maps store a distinct value for each CPU. They are useful
 // to collect metrics.
 func ExampleMap_perCPU() {
@@ -1898,6 +2108,7 @@ func ExampleMap_perCPU() {
 	if err != nil {
 		panic(err)
 	}
+	defer arr.Close()
 
 	first := []uint32{4, 5}
 	if err := arr.Put(uint32(0), first); err != nil {
@@ -1941,7 +2152,15 @@ func ExampleMap_perCPU() {
 // Note that using unsafe.Pointer is only marginally faster than
 // implementing Marshaler on the type.
 func ExampleMap_zeroCopy() {
-	hash := createHash()
+	hash, err := NewMap(&MapSpec{
+		Type:       Hash,
+		KeySize:    5,
+		ValueSize:  4,
+		MaxEntries: 10,
+	})
+	if err != nil {
+		panic(err)
+	}
 	defer hash.Close()
 
 	key := [5]byte{'h', 'e', 'l', 'l', 'o'}
@@ -1960,7 +2179,7 @@ func ExampleMap_zeroCopy() {
 	// Output: The value is: 23
 }
 
-func createHash() *Map {
+func ExampleMap_NextKey() {
 	hash, err := NewMap(&MapSpec{
 		Type:       Hash,
 		KeySize:    5,
@@ -1970,11 +2189,6 @@ func createHash() *Map {
 	if err != nil {
 		panic(err)
 	}
-	return hash
-}
-
-func ExampleMap_NextKey() {
-	hash := createHash()
 	defer hash.Close()
 
 	if err := hash.Put("hello", uint32(21)); err != nil {
@@ -2001,7 +2215,15 @@ func ExampleMap_NextKey() {
 // ExampleMap_Iterate demonstrates how to iterate over all entries
 // in a map.
 func ExampleMap_Iterate() {
-	hash := createHash()
+	hash, err := NewMap(&MapSpec{
+		Type:       Hash,
+		KeySize:    5,
+		ValueSize:  4,
+		MaxEntries: 10,
+	})
+	if err != nil {
+		panic(err)
+	}
 	defer hash.Close()
 
 	if err := hash.Put("hello", uint32(21)); err != nil {
