@@ -58,3 +58,7 @@ func TestHaveProbeReadKernel(t *testing.T) {
 func TestHaveBPFToBPFCalls(t *testing.T) {
 	testutils.CheckFeatureTest(t, haveBPFToBPFCalls)
 }
+
+func TestHaveSyscallWrapper(t *testing.T) {
+	testutils.CheckFeatureTest(t, haveSyscallWrapper)
+}
